// FUSE file system for vlbi_streamer style recordings
#define FUSE_USE_VERSION 26
#include <fuse.h>

// Project includes
#include <auto_array.h>
#include <evlbidebug.h>
#include <regular_expression.h>
#include <dosyscall.h>
#include <directory_helper_templates.h>

// Standardized C++ headers
#include <iostream>
#include <map>
#include <set>
#include <list>
#include <queue>
#include <string>
#include <cstddef>
#include <cerrno>
#include <cstring>
#include <csignal>
#include <cstdlib>
#include <climits>

// Old-style *NIX headers
#include <unistd.h>
#include <dirent.h>
#include <sys/types.h>


using namespace std;

DECLARE_EZEXCEPT(vbs_except)
DEFINE_EZEXCEPT(vbs_except)


///////////////////////////////////////////////////////////////
//
// Simple utility: scoped mutex lock
// Always comes in handy
//
///////////////////////////////////////////////////////////////
struct scoped_lock_type {
    scoped_lock_type(pthread_mutex_t* mtx):
        __m_slt_mutex_ptr( mtx )
    { ::pthread_mutex_lock(__m_slt_mutex_ptr); }

    ~scoped_lock_type() {
        ::pthread_mutex_unlock(__m_slt_mutex_ptr);
    }

    pthread_mutex_t* __m_slt_mutex_ptr;
};

///////////////////////////////////////////////////
//
//   A mountpoint monitor's private parts
//
///////////////////////////////////////////////////
struct filechunk_type;
typedef queue<filechunk_type*>  workqueue_type;

struct mphandler_data {
    bool             stop;
    bool             indexed;
    pthread_t        threadId;
    const string     mountPoint;
    workqueue_type   workqueue;
    pthread_cond_t*  condition;
    pthread_mutex_t* mutex;

    mphandler_data(string const& mp):
        stop( false ), indexed( false ), mountPoint( mp ),
        condition( new pthread_cond_t ), mutex( new pthread_mutex_t )
    {
        // Initialize the mutex and condition variable
        ::pthread_cond_init(condition, 0);
        ::pthread_mutex_init(mutex, 0);
    }
};

// Keep a global mapping of recording => filechunks
// note that we let the filechunks be an automatically sorted container
mphandler_data& get_mpdata( string const& mp );

struct filechunk_type {
    // Note: no default c'tor
    // construct from mountpoint AND full path name
    // the mountpoint will be used as lookup key for the thread to ask for the chunk's data
    filechunk_type(string const& mp, string const& fnm):
        mountPoint( mp ), pathToChunk( fnm ), chunkOffset( 0 ), chunkEnd( 0 ), __m_saved_errno( ESRCH ), __m_thread_active( false ),
        __m_bytes( 0 ), __m_cond( new pthread_cond_t ), __m_mutex( new pthread_mutex_t )
    {
        if( ::lstat(pathToChunk.c_str(), &chunkStat)<0 ) {
            DEBUG(-1, "filechunk_type[" << pathToChunk << "] - " << ::strerror(errno) << endl);
            throw errno;
        }
        chunkSize = chunkStat.st_size;

        // At this point we assume 'fnm' looks like
        // "/path/to/file/chunk.012345678"
        string::size_type   dot = fnm.find_last_of('.');

        if( dot==string::npos ) {
            DEBUG(-1, "filechunk_type[" << pathToChunk << "] - no '.' in filename!" << endl);
            throw EINVAL;
        }

        // Note: we must instruct strtoul(3) to use base 10 decoding. The
        // numbers start with a loooot of zeroes mostly and if you pass "0"
        // as third arg to strtoul(3) "accept any base, derive from prefix"
        // this throws off the automatic number-base detection
        chunkNumber = (unsigned int)::strtoul(fnm.substr(dot+1).c_str(), 0, 10);

        // Initialize the mutex and condition variable
        ::pthread_cond_init(__m_cond, 0);
        ::pthread_mutex_init(__m_mutex, 0);
    }

    // Note: declare chunkOffset as mutable such that we can later, after
    // the chunks have been sorted and put into a set, update the
    // chunkOffset value to what it should be. [Elements in a set are const
    // in order to guarantee that their sorting order is not compromised
    // as you alter the element - in this case WE know that the sorting only
    // depends on the value of 'chunkNumber' so we can safely edit
    // chunkOffset w/o worrying about compromising the set]
    string          mountPoint;
    string          pathToChunk;
    off_t           chunkSize;
    struct stat     chunkStat;
    mutable off_t   chunkOffset;
    mutable off_t   chunkEnd;
    unsigned int    chunkNumber;

#if 0
    off_t loWaterMark( void ) const {
        return chunkOffset + chunkSize/2;
    }
#endif

    bool  inRange(off_t fp) const {
        return fp>=chunkOffset && fp<chunkEnd;
        //return fp>=chunkOffset && fp<chunkOffset+chunkSize;
    }

    // Given a file pointer which must lay at or beyond the start
    // of this chunk. If pointing outside the far end, return 0
    // [this is typically true if someone seeks outside the file
    // size's range or if we've reached EOF but wait for further
    // incoming data].
    size_t remainingBytes(off_t fp) const {
        const off_t limit = chunkOffset + chunkSize;
        return (size_t)(fp>=chunkOffset && fp<=limit?(limit - fp):0);
    }

    // (pre-)reading API of the chunk.
    // It is possible to initiate reading a chunk's data and then later wait
    // for it to become available

    // This initiates a thread, reading the chunk's data, if it's not
    // already in memory. This method may safely be called multiple times,
    // although NOT from DIFFERENT THREADS at the same time on the same
    // chunk - make sure you hold the chunk lock.
    void initiateRead( int observer ) const {
        // Can used scoped lock
        scoped_lock_type    slt( __m_mutex );

        DEBUG(2, "iniateRead[" << pathToChunk << "]" << endl);

        // No observers => data isn't read yet, so must issue a request
        if( __m_observers.empty() ) {
            // Append ourselves to the appropriate workqueue - the queue
            // for the mountpoint we're located on!
            mphandler_data&   mpdata( get_mpdata(mountPoint) /*vbs_state->mpmonitorMap[mountPoint]*/ );
            scoped_lock_type  wqlock( mpdata.mutex );

            // Adding to the workqueue is not enough (guess how I found this out!)
            // we must also signal the condition - how else is the worker thread
            // to know that something happened?!
            if( !mpdata.stop ) {
                mpdata.workqueue.push( const_cast<filechunk_type*>(this) );
                ::pthread_cond_broadcast(mpdata.condition);

                // Also indicate that we're being read
                __m_thread_active = true;
            }
        }

        // Insert observer into set
        __m_observers.insert( observer );
    }

    // At some time after init you call this to wait indefinitely for the data to be
    // in memory. If it's already there it returns quickly, obviously.
    //
    // If it returns NULL, it sets errno as follows:
    //   * ESRCH  'No such process' if no call to initateRead() was made
    //     first
    //   * ENOMEM if the memory allocation for this file chunk failed
    //   * any of the errno's returned by open(2) or read(2), if a failure
    //     occurred during any of these operations, trying to read the 
    //     chunk's data into memory
    //
    unsigned char* waitForData( void ) const {
        scoped_lock_type    slt( __m_mutex );

        // Wait whilst no data & thread active.
        // As soon as either is false, we can do stuff
        while( __m_bytes==0 && __m_thread_active )
            ::pthread_cond_wait(__m_cond, __m_mutex);

        if( __m_bytes==0 )
            errno = __m_saved_errno;
        return __m_bytes;
    }

    // When the caller has no use for this chunk's data anymore, please to 
    // release it!
    void releaseData( int observer ) const {
        scoped_lock_type         slt( __m_mutex );
        const size_t             nobs   = __m_observers.size();
        observers_type::iterator obsptr = __m_observers.find( observer );

        // Check if the observer was observing this block
        if( obsptr==__m_observers.end() )
            return;

        // Remove it from the set. If the set of observers is empty, then we can safely 
        // deallocate the block.
        __m_observers.erase( obsptr );

        DEBUG(2, "filechunk[" << pathToChunk << "]::releaseData - " << pathToChunk << " data:" << (__m_bytes!=0)
                 << " nObservers:" << nobs << " -> " << __m_observers.size() << endl);

        // still observers left - we're done
        if( __m_observers.size() )
            return;

        // No more referrers?
        // Then we must at least wait for a possible thread to finish
        // [E.g. file closed before read-ahead chunk was actually needed]
        while( __m_thread_active )
            ::pthread_cond_wait(__m_cond, __m_mutex);

        // Good, now that the thread is gone, we can delete the data
        if( __m_bytes )
            ::free( __m_bytes );
        __m_bytes       = 0;
        __m_saved_errno = ESRCH;
    }

    private:
        // the mountpoint monitor thread function has
        // access to our internals
        friend void* mountpoint_monitor(void*);

        // no default c'tor!
        filechunk_type();

        // the data + mutex &cet protecting the object
        typedef std::set<int>  observers_type;

        mutable int              __m_saved_errno;
        mutable bool             __m_thread_active;
        mutable observers_type   __m_observers; 
        mutable unsigned char*   __m_bytes;
        mutable pthread_cond_t*  __m_cond;
        mutable pthread_mutex_t* __m_mutex;

        // chunk-reading function. preferrably to be executed
        // inna different thread of execution
        static void* read_chunk(void* args) {
            int              fd, e = 0;
            unsigned char*   bytes;
            filechunk_type*  fcptr = (filechunk_type*)args;

            DEBUG(2, "read_chunk[" << ::pthread_self() << "] reading " << fcptr->pathToChunk << endl);
            // If we cannot open the file ...
            if( (fd=::open(fcptr->pathToChunk.c_str(), O_RDONLY))<0 ) {
                e = errno;
                DEBUG(-1, "read_chunk[" << fcptr->pathToChunk << "] - failed to open: " << ::strerror(errno) << endl);
                ::pthread_mutex_lock(fcptr->__m_mutex);
                fcptr->__m_saved_errno = e;
                fcptr->__m_thread_active = false;
                ::pthread_cond_broadcast(fcptr->__m_cond);
                ::pthread_mutex_unlock(fcptr->__m_mutex);
                return (void*)0;
            }

            // use ::malloc() because it doesn't throw - we cannot throw
            // exceptions across threads and this executes inna different
            // thread
            if( (bytes=(unsigned char*)::malloc(fcptr->chunkSize))==0 ){
                ::close( fd );
                DEBUG(-1, "read_chunk[" << fcptr->pathToChunk << "] - insufficient memory for " << fcptr->chunkSize << " bytes" << endl);
                ::pthread_mutex_lock(fcptr->__m_mutex);
                fcptr->__m_saved_errno   = ENOMEM;
                fcptr->__m_thread_active = false;
                ::pthread_cond_broadcast(fcptr->__m_cond);
                ::pthread_mutex_unlock(fcptr->__m_mutex);
                return (void*)0;
            }

            // Nothing left than to read the data
            size_t  n = (size_t)fcptr->chunkSize;
            while( n ) {
                ssize_t nr = ::read(fd, bytes+(fcptr->chunkSize-n), n);
                if( nr<=0 ) {
                    if( nr<0 ) {
                        e = errno;
                        DEBUG(-1, "read_chunk[" << fcptr->pathToChunk << "] - failed to read() " << ::strerror(errno) << endl);
                    }
                    break;
                }
                n -= (size_t)nr;
            }
            ::close( fd );

            // Ok, indicate we're done!
            ::pthread_mutex_lock(fcptr->__m_mutex);
            fcptr->__m_saved_errno   = e;
            fcptr->__m_thread_active = false;
            fcptr->__m_bytes         = bytes;
            ::pthread_cond_broadcast(fcptr->__m_cond);
            ::pthread_mutex_unlock(fcptr->__m_mutex);

            DEBUG(2, "read_chunk[" << ::pthread_self() << "] DONE reading " << fcptr->pathToChunk << endl);
            return (void*)0;
        }
};

// Comparison operator for filechunk_type - sort by chunkNumber exclusively!
bool operator<(filechunk_type const& l, filechunk_type const& r) {
    return l.chunkNumber < r.chunkNumber;
}

// And we keep metadata per recording separately
struct metadata_type {
    off_t       recordingSize;
    time_t      lastModify; // the time stamp of the last chunk

    metadata_type():
        recordingSize( 0 ), lastModify( 0 )
    {}
};

// Have to do this before defining openfile_type ..
typedef set<filechunk_type>              filechunks_type;
typedef list<filechunks_type::iterator>  ofcache_type;

// Open file information
struct openfile_type {
    // remember last pointer to filechunk
    filechunks_type*            fileChunks;
    pthread_rwlock_t            fileRWLock;
    ofcache_type                ofCache;

    openfile_type(filechunks_type* fcptr):
        fileChunks( fcptr )
    {
        int r;
        // default mutex is good enough for us!
        if( (r=::pthread_rwlock_init(&fileRWLock, 0))!=0 )
            DEBUG(-1, "openfile_type()/pthread_rwlock_init fails - " << ::strerror(r) << endl);
    }

    // Returns iterator to chunk in cache holding the file-pointer desired
    // or fileChunks->end() if not in cache
    ofcache_type::iterator inCache(off_t offset) {
        for(ofcache_type::iterator curiterptr=ofCache.begin(); curiterptr!=ofCache.end(); curiterptr++)
            if( (*curiterptr)->inRange(offset) )
                return curiterptr;
        return ofCache.end();
    }

    // return an iterator to a block that we can start to search for for the
    // requested offset
    filechunks_type::iterator search_pos(off_t /*offset*/) const {
        // IF we have chunks in cache, return iterator to 
        // those first - the searcher knows wether to 
        // go forward or backwards from that one
        if( ofCache.begin()!=ofCache.end() )
            return *ofCache.begin();
        return fileChunks->begin();
    }

    // Add a chunk to the cache. Caller has done the appropriate locking
    // [this open file object is locked for writing AND the chunk-cache
    //  is locked]
    // If already there: no-op.
    // If not: possibly delete the entry which is the furthest behind 
    //         the chunkpos currently being inserted.
    void add_chunk(filechunks_type::iterator ptr, int fd) {
        off_t                   distance = 0;
        ofcache_type::iterator  most_distant = ofCache.end();
        DEBUG(2, "add_chunk[" << ptr->pathToChunk << "] cache size = " << ofCache.size() << endl);
        for(ofcache_type::iterator curiterptr=ofCache.begin(); curiterptr!=ofCache.end(); curiterptr++)
            if( (*curiterptr)==ptr )
                return;
            else {
                const off_t c_offs = (*curiterptr)->chunkOffset, c_dist = ptr->chunkOffset>c_offs?ptr->chunkOffset-c_offs:c_offs-ptr->chunkOffset;
                if( c_dist>distance ) {
                    distance     = c_dist;
                    most_distant = curiterptr;
                }
            }
        // not in cache yet - we already found the iterator to the block
        // that is 'most distant' from the block we're trying to insert
        // so we discard that one
        //if( ofCache.size()>3 && most_distant!=ofCache.end() ) {
        if( ofCache.size()>1 && most_distant!=ofCache.end() ) {
            DEBUG(2, "add_chunk[" << ptr->pathToChunk << "] REMOVED FROM CACHE: " << (*most_distant)->pathToChunk << endl);
            (*most_distant)->releaseData( fd );
            ofCache.erase( most_distant );
        }
        // And add the new block
        ofCache.push_back( ptr );
        DEBUG(2, "add_chunk[" << ptr->pathToChunk << "] ADDED TO CACHE" << endl);
    }
};



///////////////////////////////////////////////////
//
//   typedefs
//
///////////////////////////////////////////////////

typedef map<string, mphandler_data>     mpmonitormap_type;
typedef map<string, filechunks_type>    chunkcache_type;
typedef map<string, metadata_type>      metadatacache_type;
typedef map<int, openfile_type>         openfilecache_type;
//typedef map<string, time_t>             mountpointcheck_type;



///////////////////////////////////////////////
//   Keep state of the FUSE file system in
//   a struct of this type. It's a global
//   variable.
///////////////////////////////////////////////
const string defaultRootDir    = "/mnt";
const string defaultMk6RootDir = "/mnt/disks";

struct vbs_state_type {
    int                  readAhead;
    bool                 mark6;
    DIR*                 rootDirPtr;
    string               rootDir;
    chunkcache_type      chunkCache;
    pthread_mutex_t*     chunkMutex;
    mpmonitormap_type    mpmonitorMap;
    metadatacache_type   metadataCache;
    openfilecache_type   openfileCache;

    vbs_state_type():
        readAhead( 1 ), mark6( false ), rootDirPtr( 0 ), chunkMutex( new pthread_mutex_t )
    {
        ::pthread_mutex_init(chunkMutex, 0);
    }

    int openRootDir( void ) {
        int  rv = 0;

        if( rootDir.empty() ) {
            DEBUG(-1, "openRootDir/no root dir configured" << endl);
            return -EINVAL;
        }

        if( rootDirPtr==0 ) {
            if( (rootDirPtr=::opendir(rootDir.c_str()))==0 ) {
                rv = errno;
                DEBUG(-1, "openRootDir/failed '" << rootDir << "' - " << ::strerror(rv) << endl);
            }
        }
        return rv;
    }

    ~vbs_state_type() {
        DEBUG(2, "~vbs_state_type: stopping all monitor threads" << endl);
        // Stop all mountpoint monitors
        for( mpmonitormap_type::iterator monptr=mpmonitorMap.begin(); monptr!=mpmonitorMap.end(); monptr++) {
            mphandler_data&      mpdata = monptr->second;

            DEBUG(2, "~vbs_state_type: stopping " << monptr->first << endl);

            // Signal thread that it's time to go
            ::pthread_mutex_lock(mpdata.mutex);
            mpdata.stop = true;
            ::pthread_cond_signal(mpdata.condition);
            ::pthread_mutex_unlock(mpdata.mutex);
        }


        for( mpmonitormap_type::iterator monptr=mpmonitorMap.begin(); monptr!=mpmonitorMap.end(); monptr++) {
            mphandler_data&      mpdata = monptr->second;
            // Now join it
            int   j;
            void* dummy;

            DEBUG(2, "~vbs_state_type: joining " << monptr->first << endl);
            j = ::pthread_join(mpdata.threadId, &dummy);
            if( j!=0 ) {
                DEBUG(-1, "~vbs_state_type: failed to join [" << monptr->first << "]: " << ::strerror(errno) << endl);
            }
            DEBUG(2, "~vbs_state_type: " << monptr->first << " " << ((j==0)?("OK"):("FAIL")) << endl);
        }
        DEBUG(2, "~vbs_state_type: done" << endl);
    }
};

vbs_state_type* vbs_state = 0;


///////////////////////////////////////////////
//
//  Given a mountpoint, return a reference
//  to the mountpointhandler data
//
///////////////////////////////////////////////
mphandler_data& get_mpdata( string const& mp ) {
    mpmonitormap_type::iterator p = vbs_state->mpmonitorMap.find( mp );

    EZASSERT2( p!=vbs_state->mpmonitorMap.end(), vbs_except, EZINFO("Mountpoint " << mp << " not found in monitormap" << endl) );
    return p->second;
}


///////////////////////////////////////////////
//   Wrapper around ::pthread_create(3) -
//   creates a joinable thread with
//   ALL signals blocked.
//   Takes same arguments as ::pthread_create(3)
//   apart from the pthread_attr_t struct
///////////////////////////////////////////////
int vbs_pthread_create(pthread_t* thread, void *(*start_routine)(void*), void *arg) {
    int             pr, createrv;
    sigset_t        oldSig, newSig;
    pthread_attr_t  attr;

    // Make sure we have a joinable thread
    if( (pr=::pthread_attr_init(&attr))!=0 ) {
        DEBUG(-1, "vbs_pthread_create: pthread_attr_init fails - " << ::strerror(pr) << endl);
        return pr;
    }
    if( (pr=::pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE))!=0 ) {
        DEBUG(-1, "vbs_pthread_create: pthread_attr_setdetachstate fails - " << ::strerror(pr) << endl);
        return pr;
    }
    // Install a fully filled signal set (i.e. block all of 'm) and save the
    // current one
    if( sigfillset(&newSig)!=0 ) {
        DEBUG(-1, "vbs_pthread_create: sigfillset fails - " << ::strerror(errno) << endl);
        return errno;
    }
    if( (pr=::pthread_sigmask(SIG_SETMASK, &newSig, &oldSig))!=0 ) {
        DEBUG(-1, "vbs_pthread_create: pthread_sigmask (setting new mask) fails - " << ::strerror(pr) << endl);
        return pr;
    }
    // Now we're in a determined state and we can safely create the 
    // thread. We save the return value of this one specifically because
    // that will be the eventual return value. We do the cleanup calls
    // and if they'll fail we inform the user that but do not return *those*
    // error value(s)
    createrv = ::pthread_create(thread, &attr, start_routine, arg);
    if( createrv!=0 )
        DEBUG(-1, "vbs_pthread_create: pthread_create fails - " << ::strerror(createrv) << endl);

    // Cleanup phase: put back old signal mask & destroy pthread attributes
    if( (pr=::pthread_sigmask(SIG_SETMASK, &oldSig, 0))!=0 )
        DEBUG(-1, "vbs_pthread_create: pthread_sigmask (put back old mask) fails - " << ::strerror(pr) << endl);
    if( (pr=::pthread_attr_destroy(&attr))!=0 )
        DEBUG(-1, "vbs_pthread_create: pthread_attr_destroy fails - " << ::strerror(pr) << endl);

    // Phew. Finally done.
    return createrv;
}

///////////////////////////////////////////////
//  The mountpoint monitor function.
//  It is a thread function
///////////////////////////////////////////////

// Operator that monitors directories inside
//   <rootDir>/diskN/
// to see if they've changed. If they've changed
// then likely there are files added or deleted.
//
// It returns true IF
//    * entry is a dir AND
//    * entry has r,x such that we can both enter into the dir and read it AND
//    * entry is m_time > requested time stamp
#define RX_USER(m) ((m & S_IRUSR) && (m & S_IXUSR))
struct changedSince {
    struct status_type {
        // 'maybeRecording' is to be read as
        // "the entry is possibly a recording"
        // (thus NOT: "the entry may currently be busy recording")
        bool   maybeRecording;
        bool   hasChanged;
        string fileName;

        status_type():
            maybeRecording( false ), hasChanged( false )
        {}

        status_type(bool isDir, bool changed, string const& fn):
            maybeRecording( isDir ), hasChanged( changed ), fileName( fn )
        {}
    };

    typedef status_type  value_type;

    changedSince(time_t ts):
        __m_timestamp( ts )
    {}

    value_type operator()(string const& fnm) const {
        // Check if it's a directory and wether it's been changed since the
        // last check
        struct stat             s;
        const string::size_type slash = fnm.find_last_of("/");

        if( ::lstat(fnm.c_str(), &s)<0 )
            return status_type();

        // We're only interested in directories that have changed since we
        // last checked. Be implicitly ignore the 'lost+found' directory, as
        // it is not a potential recording ...
        // We must have r,x access to the directory [in order to descend into it]
        return status_type(S_ISDIR(s.st_mode) && RX_USER(s.st_mode) && fnm.find("lost+found")==string::npos,
                           (s.st_mtime>__m_timestamp),
                           fnm.substr(slash==string::npos?0:slash+1));
    }

    private:
        // no default c'tor possible!
        changedSince();

        time_t  __m_timestamp;
};

struct isRecordingChunk {
    isRecordingChunk(string const& recname):
        __m_regex( string("^")+recname+"\\.[0-9]{8}$" )
    {}

    bool operator()(string const& entry) const {
        DEBUG(4, "checking entry " << entry << " against " << __m_regex.pattern() << endl);
        return __m_regex.matches(entry);
    }

    Regular_Expression  __m_regex;

    private:
        isRecordingChunk();
};

// note that we only ever ADD files to the chunk cache. If someone deletes
// the backing storage from out under us, that's too bad ...
void* mountpoint_monitor(void* args) {
    typedef dir_mapper<changedSince>  changeChecker_type;
    typedef changeChecker_type::value_type checkresult_type;

    // We get passed a pointer to string
    DIR*            dirp = 0;
    bool            except = false;
    mphandler_data* mpdata = (mphandler_data*)args;

    if( mpdata==0 ) {
        DEBUG(-1, "mountpoint_monitor/null-pointer passed as handler data!" << endl);
        return (void*)0;
    }
    DEBUG(2, "mountpoint_monitor[" << mpdata->mountPoint << "] starting" << endl);

    try {
        // Let's check that mother!
        checkresult_type status = changeChecker_type( changedSince(0) )( mpdata->mountPoint );
       
        // Loop over all directories and analyze the ones that have changed
        for(checkresult_type::const_iterator ptr=status.begin(); ptr!=status.end(); ptr++) {
            changedSince::value_type const&  entry = ptr->second;

            // We don't even _try_ to look inside 'lost+found'
            if( entry.maybeRecording || entry.hasChanged ) {
                DEBUG(3, "[" << mpdata->mountPoint << "] " << ptr->first << " {" << entry.fileName << "}" << endl);
                // inspect files inside the directory - if they match
                // "<directoryname>.XXXXXXXX  [X = 0..9] then it is a chunk
                // of a flexbuff recording

                if( (dirp=::opendir(ptr->first.c_str()))==0 ) {
                    DEBUG(1, "mountpoint_monitor[" << mpdata->mountPoint << "] opendir " << ptr->first << " fails - " << ::strerror(errno) << endl);
                    continue;
                }
                direntries_type   chunks = dir_filter(dirp, isRecordingChunk(entry.fileName));
                scoped_lock_type  chunklock(vbs_state->chunkMutex);

                for(direntries_type::const_iterator p=chunks.begin(); p!=chunks.end(); p++)
                    vbs_state->chunkCache[entry.fileName].insert( filechunk_type(mpdata->mountPoint, ptr->first+"/"+*p) );

                ::closedir(dirp);
                dirp = 0;
            }
        }

        // Signal that we're done indexing
        ::pthread_mutex_lock(mpdata->mutex);
        mpdata->indexed = true;
        ::pthread_cond_signal(mpdata->condition);
        ::pthread_mutex_unlock(mpdata->mutex);

        // And drop into our inner loop!
        while( true ) {
            bool            stop    = false; // copies of shared state variables
            filechunk_type* workptr = 0;

            // lock self & condition wait for something to happen
            ::pthread_mutex_lock(mpdata->mutex);
            while( mpdata->stop==false && mpdata->workqueue.size()==0 )
                ::pthread_cond_wait(mpdata->condition, mpdata->mutex);

            stop = mpdata->stop;
            if( mpdata->workqueue.size() ) {
                workptr = mpdata->workqueue.front();
                mpdata->workqueue.pop();
            }
            ::pthread_mutex_unlock(mpdata->mutex);

            // Check what the message was
            if( stop==true )
                break;

            if( workptr==0 ) {
                DEBUG(-1, "mountpoint_monitor[" << mpdata->mountPoint << "] got nullptr in workqueue?!" << endl);
                continue;
            }
            // Crap. There's no denying anymore - we have to Do Work(tm)
            filechunk_type::read_chunk( workptr );
        }
        // Ok, we're not reading anymore. Inform all queued bloccks
        // that it ain't gonna happen no more
        ::pthread_mutex_lock(mpdata->mutex);
        while( mpdata->workqueue.size() ) {
            filechunk_type* fcptr = mpdata->workqueue.front();

            ::pthread_mutex_lock(fcptr->__m_mutex);
            fcptr->__m_thread_active = false;
            ::pthread_cond_broadcast(fcptr->__m_cond);
            ::pthread_mutex_unlock(fcptr->__m_mutex);
            mpdata->workqueue.pop();
        }
        ::pthread_mutex_unlock(mpdata->mutex);
    }
    catch( int eno ) {
        except = true;
        DEBUG(-1, "mountpoint_monitor[" << mpdata->mountPoint << "] caught errno " << eno << " - " << ::strerror(eno) << endl);
    }
    catch( std::exception& e ) {
        except = true;
        DEBUG(-1, "mountpoint_monitor[" << mpdata->mountPoint << "] caught exception " << e.what() << endl);
    }
    catch( ... ) {
        except = true;
        DEBUG(-1, "mountpoint_monitor[" << mpdata->mountPoint << "] caught unknown exception" << endl);
    }
    if( except ) {
        // Signal to main loop that (1) we're done initializing and (2)
        // that we don't accept any requests
        ::pthread_mutex_lock(mpdata->mutex);
        mpdata->indexed = true;
        mpdata->stop    = true;
        ::pthread_cond_signal(mpdata->condition);
        ::pthread_mutex_unlock(mpdata->mutex);
    }
    DEBUG(2, "mountpoint_monitor[" << mpdata->mountPoint << "] stopping" << endl);
    return (void*)0;
}


// Predicate for a directory entry to be a mountpoint
// 1. name must be "disk[0-9]+"
// 2. must refer to a directory
struct isMountpoint {
    bool operator()(string const& entry) const {
        Regular_Expression      rxDisk("^disk[0-9]{1,}$");
        struct stat             status;
        string::size_type       slash = entry.find_last_of("/");

        // IF there is a slash, we skip it, if there isn't, we
        // use the whole string
        if( slash==string::npos )
            slash = 0;
        else
            slash += 1;
        DEBUG(3, "isMountpoint: checking name " << entry.substr(slash) << endl);
        if( !rxDisk.matches(entry.substr(slash)) )
            return false;
        if( ::lstat(entry.c_str(), &status)<0 ) {
            DEBUG(2, "predMountpoint: ::lstat fails on " << entry << " - " << ::strerror(errno) << endl);
            return false;
        }
        // We must have r,x access to the directory [in order to descend into it]
        return S_ISDIR(status.st_mode) && (status.st_mode & S_IRUSR) && (status.st_mode & S_IXUSR);
    }
};


///////////////////////////////////////////////
//
//   Here follow the implementations 
//   of the file system functions
//   that we support
//
///////////////////////////////////////////////


void* vbs_init(struct fuse_conn_info* /*conn*/) {
    DEBUG(-1, "vbs_init: starting" << endl);
    try {
        direntries_type    mountPoints;
        mpmonitormap_type& mpmon( vbs_state->mpmonitorMap );

        // Depending on Mk6 or not we do things differently
        if( vbs_state->mark6 ) {
            // Just push all Mark6 mount points
            for(unsigned int mod=1; mod<5; mod++) {
                for(unsigned int dsk=0; dsk<8; dsk++) {
                    struct stat   st;
                    ostringstream oss;
                    oss << vbs_state->rootDir << "/" << mod << "/" << dsk;
                    if( ::stat(oss.str().c_str(), &st)<0 ) {
                        DEBUG(-1, "vbs_init: failed to stat " << oss.str() << endl);
                        continue;
                    }
                    // Is it a directory? 
                    if( S_ISDIR(st.st_mode) )
                        mountPoints.insert( oss.str() );
                }
            }
        } else {
            // Check rootDir for matching mount points
            mountPoints = dir_filter(vbs_state->rootDir, isMountpoint());
        }

        EZASSERT2( mountPoints.size()>0, vbs_except, EZINFO("No mountpoints found in " << vbs_state->rootDir) );

        for( direntries_type::const_iterator mp=mountPoints.begin(); mp!=mountPoints.end(); mp++) {
            int                                      r;
            pair<mpmonitormap_type::iterator, bool>  insres = mpmon.insert( make_pair(*mp, mphandler_data(*mp)) );

            DEBUG(2, "vbs_init: Found mountpoint " << *mp << endl);
            EZASSERT2(insres.second==true, vbs_except, EZINFO("Duplicate entry for mountpoint " << *mp));

            // Start the monitor thread
            if( (r=vbs_pthread_create(&insres.first->second.threadId, mountpoint_monitor, &insres.first->second))!=0 )
                DEBUG(-1, "vbs_init: Failed to create thread!" << endl);
        }

        // For each thread, wait for it to have signalled it's done initializing
        DEBUG(-1, "vbs_init: please wait whilst indexing " << mpmon.size() << " mountpoints" << endl);
        for( mpmonitormap_type::iterator monptr=mpmon.begin(); monptr!=mpmon.end(); monptr++) {
            bool                 stop;
            mphandler_data&      mpdata = monptr->second;

            ::pthread_mutex_lock(mpdata.mutex);
            while( mpdata.stop==false && mpdata.indexed==false )
                ::pthread_cond_wait(mpdata.condition, mpdata.mutex);
            stop = mpdata.stop;
            ::pthread_mutex_unlock(mpdata.mutex);
            DEBUG(1, "vbs_init: " << monptr->first << " " << ((stop)?("EXIT"):("OK")) << endl);
        }
        DEBUG(-1, "vbs_init: indexing " << mpmon.size() << " mountpoints finished" << endl);

        DEBUG(-1, "vbs_init: indexing metadata of " << vbs_state->chunkCache.size() << " recordings" << endl);
        // Now we update the file chunks with their respective filepointer start/end values
        // and the metadata
        for(chunkcache_type::iterator recptr=vbs_state->chunkCache.begin(); recptr!=vbs_state->chunkCache.end(); recptr++) {
            off_t             filepos   = 0;
            metadata_type&    metadata = vbs_state->metadataCache[recptr->first];
            filechunks_type&  recchunks = recptr->second;

            for(filechunks_type::iterator ccptr=recchunks.begin(); ccptr!=recchunks.end(); ccptr++) {
                ccptr->chunkOffset   = filepos;
                filepos             += ccptr->chunkSize;
                ccptr->chunkEnd      = filepos;
                metadata.lastModify  = max(metadata.lastModify, ccptr->chunkStat.st_mtime);
            }
            metadata.recordingSize = filepos; 
        }
        DEBUG(-1, "vbs_init: indexing metadata finished" << endl);
    }
    catch( int erno ) {
        DEBUG(-1, "vbs_init/caught deadly errno - " << ::strerror(erno) << endl);
    }
    catch( const exception& e ) {
        DEBUG(-1, "vbs_init/caught deadly exception - " << e.what() << endl);
    }
    catch( ... ) {
        DEBUG(-1, "vbs_init/caught deadly unknown exception" << endl);
    }
    return (void*)vbs_state;
}

int vbs_readdir(char const* path, void* buf, fuse_fill_dir_t filler, off_t /*offset*/, struct fuse_file_info* /*fi*/) {
    DEBUG(1, "vbs_readdir[" << path << "]" << endl);
    // We will *never* allow subdirectories so any request for anything
    // other than "/" (the root of this file system) we're going to return
    // an error!
    if( string(path)!="/" )
        return -ENOENT;
    // Now fill the directory listing
    filler(buf, ".",  0, 0);
    filler(buf, "..", 0, 0);
    // Grab the lock on the chunkCache and fetch all the keys - they are the
    // recording names!
    ::pthread_mutex_lock( vbs_state->chunkMutex );

    const chunkcache_type&  chunks = vbs_state->chunkCache;
    for(chunkcache_type::const_iterator p=chunks.begin(); p!=chunks.end(); p++)
        filler(buf, p->first.c_str(), 0, 0);

    ::pthread_mutex_unlock( vbs_state->chunkMutex );
    return 0;
}


int vbs_getattr(char const* path, struct stat* statbuf) {
    const mode_t defaultMode = S_IRWXU | S_IRWXG | S_IRWXO;
    const time_t now = ::time(0);
    DEBUG(1, "vbs_getattr[" << path << "]" << endl);
    // before we create the recording name out of the path,
    // we have to strip leading slashes!
    while( *path=='/' ) path++;
    const string recname( path );
  
    // If someone requests the attrs of "/" it's the props of
    // the FUSE mountpoint itself, so we list it as a directory 
    // [note, after stripping leading slashes, the root path
    //  becomes the empty string!]
    if( recname=="" ) {
        statbuf->st_mode  = S_IFDIR | defaultMode;
        statbuf->st_mtime = now;
        statbuf->st_ctime = now;
        statbuf->st_atime = now;
        return 0;
    }
    const chunkcache_type&             chunkCache = vbs_state->chunkCache;
    const metadatacache_type&          metadataCache = vbs_state->metadataCache;
    chunkcache_type::const_iterator    recptr;
    metadatacache_type::const_iterator metaptr;

    // Do lookup of recording whilst we hold the mutex
    // [and compute file size]
    ::pthread_mutex_lock(vbs_state->chunkMutex);
    recptr  = chunkCache.find( recname );
    metaptr = metadataCache.find( recname );
    ::pthread_mutex_unlock(vbs_state->chunkMutex);

    if( recptr==chunkCache.end() || metaptr==metadataCache.end() ) {
        DEBUG(-1, "vbs_getattr[" << path << "] - not found in cache?!!!" << endl);
        return -ENOENT;
    }
    // It's an existing recording!
    const metadata_type&    meta = metaptr->second;
    statbuf->st_mode  = S_IFREG | defaultMode;
    statbuf->st_size  = meta.recordingSize;
    statbuf->st_mtime = meta.lastModify;
    statbuf->st_ctime = meta.lastModify;
    statbuf->st_atime = meta.lastModify;
    return 0;
}

void vbs_destroy(void* ptr) {
    DEBUG(1, "vbs_destroy" << endl);
    delete (vbs_state_type*)ptr;
    vbs_state = 0;
}

#if 0
int vbs_create(char const* path, mode_t, struct fuse_file_info*) {
    DEBUG(1, "vbs_create[" << path << "]" << endl);
    return -EPERM;
}
#endif

int vbs_open(char const* path, struct fuse_file_info* fi) {
    static int          nopen = 0;
    //openfilecache_type& ofc = vbs_state->openfileCache;

    // Skip leading slash(es) in path
    while( *path=='/' )
        path++;
    int                       res = 0;
    const string              recname( path );
    chunkcache_type&          chunkcache = vbs_state->chunkCache;
    openfilecache_type&       openfiles  = vbs_state->openfileCache;
    chunkcache_type::iterator recchunksptr;

    ::pthread_mutex_lock(vbs_state->chunkMutex);
    if( (recchunksptr=chunkcache.find(recname))==chunkcache.end() ) {
        res = -ENOENT;
    } else {
        //fi->fh = INT_MAX - (int)ofc.size();
        fi->fh = INT_MAX - nopen++;
        openfiles.insert( make_pair(fi->fh, openfile_type(&recchunksptr->second)) );
    }
    ::pthread_mutex_unlock(vbs_state->chunkMutex);
    DEBUG(1, "vbs_open[" << path << "] res=" << res << " fh=" << fi->fh << endl);

    return res;
}

int vbs_read(char const* path, char* buff, size_t n, off_t offset, struct fuse_file_info* fi) {
    openfilecache_type::iterator ofp = vbs_state->openfileCache.find( fi->fh );

    if( ofp==vbs_state->openfileCache.end() )
        return -ENOENT;

    static volatile unsigned int  nrd = 0;
    const unsigned int rd = nrd++;

    int              eno  = 0;
    bool             fail = false, atEnd = false;
    const size_t     norg = n;
    openfile_type&   of = ofp->second;
    filechunks_type& chunks = *of.fileChunks;

    DEBUG(3, "vbs_read[" << path << ", " << rd << ", " << ::pthread_self() << "] offset=" << offset << " n=" << n << endl);
    while( n && !fail && !atEnd ) {
        bool                      incache;
        size_t                    navail = (size_t)-1;
        ofcache_type::iterator    cacheentry;
        filechunks_type::iterator ptr;

        // Ask the openfile cache if the current offset is somewhere in one 
        // of the cached blocks
        ::pthread_rwlock_rdlock(&of.fileRWLock);
        
        cacheentry = of.inCache(offset);
        incache    = (cacheentry!=of.ofCache.end() && (navail=(*cacheentry)->remainingBytes(offset))!=0);

        // If offset not found in the cache, at least ask it
        // for a reasonable starting point for our search.
        // We MUST lock the chunkMutex because the "search_pos()"
        // does access the chunk-list ... which _may_ be modified
        // whilst we're trying to read it ...
        if( !incache ) {
            ::pthread_mutex_lock(vbs_state->chunkMutex);
            ptr = of.search_pos( offset );
            ::pthread_mutex_unlock(vbs_state->chunkMutex);
        } else {
            // We have to do the memcpy() whilst we have the read-lock on
            // the cache; if someone behind our backs would remove the block
            // from the cache the data would be gone!
            size_t          nr = min(n, navail);
            unsigned char*  data;

            ptr  = (*cacheentry);
            data = ptr->waitForData();
            if( !(fail=(data==0)) ) {
                ::memcpy(buff, data+(offset - ptr->chunkOffset), nr);

                buff   += nr;
                n      -= nr;
                offset += nr;
            } else {
                eno = errno;
                DEBUG(-1, "vbs_read: " << ptr->pathToChunk << " FAILS: " << ::strerror(errno) << endl);
            }
        }

        ::pthread_rwlock_unlock(&of.fileRWLock);

        // navail>0 implies we did copy some data - we may have to do some
        // more
        if( fail || incache )
            continue;

        // Requested offset was not found in the cache.
        // So now we has to go look for a block which *does*
        // contain the requested offset!

        // We must go search for the next chunk (do lock the chunk table
        // for that - no mods whilst we're scanning!)
        // the 'ptr' variable has already been filled in with an appropriate
        // starting point for our search
        ::pthread_mutex_lock(vbs_state->chunkMutex);

        while( ptr!=chunks.end() && !ptr->inRange(offset) ) {
            if( offset>=ptr->chunkEnd )
                ptr++;
            else if( ptr!=chunks.begin() )
                ptr--;
            else
                break;
        }
        // Whilst we have the lock, do issue read-aheads
        // [we don't need the lock for the read-aheads
        // but we do for checking against ".end()"!]
        if( (atEnd = (ptr==chunks.end()))==false ) {
            // Issue read requests for this block and the next N ones
            int                       ra = vbs_state->readAhead;
            filechunks_type::iterator raptr = ptr;
            while( ra>=0 && raptr!=chunks.end() ) {
                raptr->initiateRead( ofp->first );
                raptr++;
                ra--;
            }
        }
        ::pthread_mutex_unlock(vbs_state->chunkMutex);

        // Ok, we have found a chunk that we need in the cache!
        // do insert it
        if( !atEnd ) {
            ::pthread_rwlock_wrlock(&of.fileRWLock);
            of.add_chunk( ptr, ofp->first );
            ::pthread_rwlock_unlock(&of.fileRWLock);
        }
    }
    DEBUG(3, "vbs_read[" << path << ", " << rd << ", " << ::pthread_self() << "] offset=" << offset << " n=" << n << " errno=" << eno << " FINISHED" << endl);

    if( eno )
        return -eno;
    return (norg - n);
}

int vbs_release(char const* path, struct fuse_file_info* fi) {
    ::pthread_mutex_lock(vbs_state->chunkMutex);

    openfilecache_type::iterator  ofp = vbs_state->openfileCache.find( fi->fh );
    DEBUG(1, "vbs_release[" << path << "] fh=" << fi->fh << endl);

    if( ofp==vbs_state->openfileCache.end() ) {
        ::pthread_mutex_unlock(vbs_state->chunkMutex);
        return -ENOENT;
    }

    openfile_type&            of = ofp->second;

    // Wait for issued chunk-read-requests to finish and properly clean them
    // up
    for(filechunks_type::iterator fcptr = of.fileChunks->begin(); fcptr!=of.fileChunks->end(); fcptr++) {
        fcptr->waitForData();
        fcptr->releaseData( ofp->first );
    }
    DEBUG(1, "vbs_release[" << path << "] done" << endl);
    vbs_state->openfileCache.erase( ofp );

    ::pthread_mutex_unlock(vbs_state->chunkMutex);
    return 0;
}
///////////////////////////////////////////////
//   A filesystem must define a set
//   of operations with callback functions
//   on the file system objects it
//   represents. All callbacks are 
//   registered in the fuse_operations struct
///////////////////////////////////////////////
struct vbs_fuse_operations :
    public fuse_operations {

        vbs_fuse_operations() {
            init    = &vbs_init;
            destroy = &vbs_destroy;
            getdir  = NULL;
            readdir = &vbs_readdir;
#if 0
            create  = &vbs_create;
#endif
            getattr = &vbs_getattr;
            open    = &vbs_open;
            read    = &vbs_read;
            release = &vbs_release; // called on close(2)
        }
};

vbs_fuse_operations vbs_oper;
#if 0
struct fuse_operations vbs_oper = {
    //.getattr = vbs_getattr,
    //.readdir = vbs_readdir,
    //.open = vbs_open,
    //.read = vbs_read,
    //.readlink = vbs_readlink,
    // no .getdir -- that's deprecated
    //.getdir = NULL,
    //.mknod = vbs_mknod,
    //.mkdir = vbs_mkdir,
    //.unlink = vbs_unlink,
    //.rmdir = vbs_rmdir,
    //.symlink = vbs_symlink,
    //.rename = vbs_rename,
    //.link = vbs_link,
    //.chmod = vbs_chmod,
    //.chown = vbs_chown,
    //.truncate = vbs_truncate,
    //.utime = vbs_utime,
    //.write = vbs_write,
    /** Just a placeholder, don't set */ // huh???
    //.statfs = vbs_statfs,
    //.flush = vbs_flush,
    //.release = vbs_release,
    //.fsync = vbs_fsync,
    //.setxattr = vbs_setxattr,
    //.getxattr = vbs_getxattr,
    //.listxattr = vbs_listxattr,
    //.removexattr = vbs_removexattr,
    //.opendir = vbs_opendir,
    //.releasedir = vbs_releasedir,
    //.fsyncdir = vbs_fsyncdir,
    .init = vbs_init,
    .destroy = vbs_destroy,
    //.access = vbs_access,
    //.create = vbs_create,
    //.ftruncate = vbs_ftruncate,
};
#endif


///////////////////////////////////////////////
//   Command line handling
//   There's only a few options
//   that we support
//    -m <int>   setting debug level
//    -r <path>  set root dir for disks
//    -f         FUSE 'foreground' flag
///////////////////////////////////////////////
struct vbs_settings {
    int          read_ahead;
    int          debug_level;
    bool         help;
    bool         mark6;
    char const*  root_dir;

    // Initialize with defaults
    vbs_settings():
        read_ahead( 0 ), debug_level( 0 ), help( false ), mark6( false ), root_dir( 0 )
    {}
};

ostream& operator<<(ostream& os, const vbs_settings& vbo) {
    os << "-m " << vbo.debug_level << " ";
    if( vbo.root_dir )
        os << "-r " << vbo.root_dir << " ";
    if( vbo.read_ahead>0 )
        os << "-n " << vbo.read_ahead << " ";
    if( vbo.help )
        os << "-h ";
    if( vbo.mark6 )
        os << "-6 ";
    return os;
}

#define OFFSETOF(tp, fld) \
    (size_t)((unsigned char*)(&(((tp*)1)->fld)) - (unsigned char*)1)

//#define VBSFS_OPT(t, p, v) { t, offsetof(struct vbs_settings, p), v }
#define VBSFS_OPT(t, p, v) { t, OFFSETOF(struct vbs_settings, p), v }
struct fuse_opt vbs_options[] = {
    VBSFS_OPT("-m %i", debug_level, 0),
    VBSFS_OPT("-n %i", read_ahead,  0),
    VBSFS_OPT("-r %s", root_dir,    0),
    VBSFS_OPT("-6"   , mark6      , 1),
    //{ "-6", -1, 0},
    // The FUSE_OPT_KEY() ones are handled
    // by vbs_option_proc()
    FUSE_OPT_END
};

int vbs_option_proc(void* data, const char* arg, int /*key*/, struct fuse_args* /*outargs*/) {
    int           rv = 1;
    const string  arg_s( arg );
    vbs_settings* settingsptr = (vbs_settings*)data;

    DEBUG(2, "vbs_option_proc: processing " << arg_s << endl);
    if( arg_s=="-h" )
        settingsptr->help  = true;
    if( arg_s=="-6" )
        settingsptr->mark6 = true;
    return rv;
}

void Usage( void ) {
    cout << "VBS specific command line arguments:" << endl << endl
         << "   -m <int>    set VBS_FS debug print level to <n>" << endl
         << "               higher <n> => more output, default 0 (no output)" << endl
         << "   -n <int>    issue <int> chunks read-ahead. Only useful if files" << endl
         << "               are read in strict linear fashion, otherwise it will" << endl
         << "               sigificantly HURT performance (default '1')" << endl
         << "   -r <path>   path to vbs_fs 'root' directory; assume all disks" << endl
         << "               are mounted under <path>/disk<n> with <n> being" << endl
         << "               a non-negative integer" << endl
         << "   -6          Use Mark6 disk mount points" << endl;
    return;
}



///////////////////////////////////////////////
//
//         The main function
//
///////////////////////////////////////////////
int main(int argc, char** argv) {
    // Before doing _anything_ check if we're not root/suid root
    if( ::getuid()==0 || ::geteuid()==0 ) {
        DEBUG(-1, "Running vbs_fuse as root opens unnacceptable security holes" << endl);
        return -11;
    }

    // Now we can safely start parsing
    int                 fuse_stat;
    vbs_settings        settings;
    struct fuse_args    args = FUSE_ARGS_INIT(argc, argv);

    vbs_state = new vbs_state_type();

    // Do command line parsing
    ::fuse_opt_parse(&args, (void*)&settings, &vbs_options[0], vbs_option_proc);
    dbglev_fn( settings.debug_level );
    DEBUG(1, "fuse_opt_arg done [" << settings << "]" << endl);

    if( settings.help )
        Usage();

    // If no root dir given, default to "/mnt"
    if( settings.help==false && settings.root_dir==0 )
        DEBUG(-1, "WARN: No root dir specified for disks, defaulting to " << 
                  (settings.mark6 ? defaultMk6RootDir+"/[1234]/[0-7]" : defaultRootDir+"/disk[0-9]+") << endl);
    vbs_state->mark6   = settings.mark6;
    vbs_state->rootDir = (settings.root_dir==0?
                          (vbs_state->mark6?defaultMk6RootDir:defaultRootDir):
                          string(settings.root_dir));

    // Set read-ahead value - only if it's >0
    if( settings.read_ahead<0 ) {
        DEBUG(-1, "invalid value " << settings.read_ahead << " for read ahead!" << endl);
        return -EINVAL;
    }
    if( settings.read_ahead>0 )
        vbs_state->readAhead = settings.read_ahead;

    // Before init, do open the rootdir - (no access(2) or stat(2) because
    // that means that between access/stat and the open the dir may have
    // gone, its permissions altered or god knows what
    if( const int r = vbs_state->openRootDir() )
        return r;

    // Now it's about time to drop into FUSE's main loop
    DEBUG(3, "dropping into fuse_main" << endl);
    fuse_stat = fuse_main(args.argc, args.argv, &vbs_oper, vbs_state);

    DEBUG(3, "fuse_stat returned " << fuse_stat << endl);
    return fuse_stat;
}
